@@ -39,13 +39,14 @@ def auto_numbered(field_name):
     return not re.match(r"[A-Za-z0-9_]", field_name)
 
 
-def process_slot_value(name, value, values):
+def process_slot_value(name, value, values, placeholders):
     if isinstance(value, Fragment):
         return value
     else:
-        placeholder = Placeholder(name)
-        values[placeholder] = value
-        return placeholder
+        if name not in placeholders:
+            placeholders[name] = Placeholder(name)
+            values[placeholders[name]] = value
+        return placeholders[name]
 
 
 @dataclasses.dataclass
@@ -65,7 +66,7 @@ class Fragment:
             elif isinstance(part, Placeholder):
                 parts.append(part)
                 values[part] = self.values[part]
-            elif slots and isinstance(part, Slot):
+            elif slots is not None and isinstance(part, Slot):
                 slot_value = slots[part.name]
                 if isinstance(slot_value, Fragment):
                     slot_value.flatten_into(parts, values, slots)
@@ -79,7 +80,6 @@ class Fragment:
     def compile(self) -> Callable[..., "Fragment"]:
         flattened = self.flatten()
         env = dict(
-            parts=flattened.parts,
             in_values=flattened.values,
             process_slot_value=process_slot_value,
             Fragment=Fragment,
@@ -87,15 +87,19 @@ class Fragment:
         func = [
             "def compiled(**slots):",
             " values = in_values.copy()",
+            " placeholders = {}",
             " return Fragment([",
         ]
         for i, part in enumerate(flattened.parts):
             if isinstance(part, Slot):
                 func.append(
-                    f"  process_slot_value({repr(part.name)}, slots[{repr(part.name)}], values),"
+                    f"  process_slot_value({repr(part.name)}, slots[{repr(part.name)}], values, placeholders),"
                 )
+            elif isinstance(part, str):
+                func.append(f"  {repr(part)},")
             else:
-                func.append(f"  parts[{i}],")
+                env[f"part_{i}"] = part
+                func.append(f"  part_{i},")
         func += [" ], values)"]
         exec("\n".join(func), env)
         return env["compiled"]  # type: ignore
@@ -163,8 +167,11 @@ class SQLFormatter:
                 if auto_numbered(field_name):
                     field_name = f"{next_auto_field}{field_name}"
                     next_auto_field += 1
-                value = fmtr.get_field(field_name, args, kwargs)[0]
-                if isinstance(value, Fragment):
+                try:
+                    value = fmtr.get_field(field_name, args, kwargs)[0]
+                except KeyError:
+                    value = Slot(field_name)
+                if isinstance(value, Fragment) or isinstance(value, Slot):
                     parts.append(value)
                 else:
                     if field_name not in placeholders:
