@@ -86,6 +86,7 @@ def column_info_for_field(field):
 
 
 T = TypeVar("T", bound="ModelBase")
+U = TypeVar("U")
 
 
 class ModelBase(Mapping[str, Any]):
@@ -104,7 +105,7 @@ class ModelBase(Mapping[str, Any]):
             cls.primary_key_names = tuple(primary_key)
 
     @classmethod
-    def _cached(cls, key: tuple, thunk: Callable[[], Any]) -> Any:
+    def _cached(cls, key: tuple, thunk: Callable[[], U]) -> U:
         try:
             fragment_cache: Dict[tuple, Any] = cls._fragment_cache  # type: ignore
         except AttributeError:
@@ -228,12 +229,11 @@ class ModelBase(Mapping[str, Any]):
                 "SELECT {fields} FROM {name} WHERE {where}",
                 fields=sql.list(cls.field_names_sql()),
                 name=cls.table_name_sql(),
-                where=sql.slot("where"),
             ).compile(),
         )
         query = cached(where=where)
         if for_update:
-            query = sql("{} FOR UPDATE", query)
+            query = Fragment([query, " FOR UPDATE"])
         return query
 
     @classmethod
@@ -272,10 +272,15 @@ class ModelBase(Mapping[str, Any]):
         return cls(**row)  # type: ignore
 
     def insert_sql(self, exclude: FieldNamesSet = ()) -> Fragment:
-        return sql(
-            "INSERT INTO {table} ({fields}) VALUES ({values})",
-            table=self.table_name_sql(),
-            fields=sql.list(self.field_names_sql(exclude=exclude)),
+        cached = self._cached(
+            ("insert_sql", tuple(sorted(exclude))),
+            lambda: sql(
+                "INSERT INTO {table} ({fields}) VALUES ({values})",
+                table=self.table_name_sql(),
+                fields=sql.list(self.field_names_sql(exclude=exclude)),
+            ).compile(),
+        )
+        return cached(
             values=sql.list(self.field_values_sql(exclude=exclude, default_none=True)),
         )
 
@@ -317,7 +322,6 @@ class ModelBase(Mapping[str, Any]):
                 "DELETE FROM {table} WHERE ({pks}) IN (SELECT * FROM {unnest})",
                 table=cls.table_name_sql(),
                 pks=sql.list(sql.identifier(pk) for pk in cls.primary_key_names),
-                unnest=sql.slot("unnest"),
             ).compile(),
         )
         return cached(
@@ -339,7 +343,6 @@ class ModelBase(Mapping[str, Any]):
                 "INSERT INTO {table} ({fields}) SELECT * FROM {unnest}",
                 table=cls.table_name_sql(),
                 fields=sql.list(cls.field_names_sql()),
-                unnest=sql.slot("unnest"),
             ).compile(),
         )
         return cached(
